try:
    from lingua import Language, LanguageDetectorBuilder
    has_lingua = True

    languages = [Language.ENGLISH, Language.FRENCH, Language.GERMAN, Language.SPANISH, Language.ITALIAN,
                 Language.PORTUGUESE, Language.DUTCH, Language.DANISH, Language.SWEDISH, Language.NYNORSK,
                 Language.FINNISH, Language.POLISH, Language.UKRAINIAN]
    lang_detector = LanguageDetectorBuilder.from_languages(*languages).build()
except ImportError:
    has_lingua = False


def indentify_language(text: str) -> str | None:
    return lang_detector.detect_language_of(text).iso_code_639_1.name.lower()



import io

from bpm_ai_core.speech_recognition.asr import ASRModel
from bpm_ai_core.util.audio import load_audio

try:
    from faster_whisper import WhisperModel
    has_faster_whisper = True
except ImportError:
    has_faster_whisper = False


class FasterWhisperASR(ASRModel):
    """
    Local `OpenAI` Whisper Automatic Speech Recognition (ASR) model for transcribing audio.

    To use, you should have the ``faster_whisper`` python package installed.
    """

    def __init__(
        self,
        model_size: str = "small",
        device: str = "cpu",
        compute_type: str = "int8"
    ):
        if not has_faster_whisper:
            raise ImportError('faster_whisper is not installed')
        self.model_size = model_size
        self.model = WhisperModel(
            model_size,
            device=device,
            compute_type=compute_type
        )

    async def _transcribe(self, audio: str | io.BytesIO, language: str = None) -> str:
        if isinstance(audio, str):
            audio = load_audio(audio)
        segments, info = self.model.transcribe(audio, language=language)
        return "".join([s.text for s in list(segments)])



import io
from abc import ABC, abstractmethod
from typing import Optional, Union

from bpm_ai_core.tracing.tracing import Tracing


class ASRModel(ABC):
    """
    Automatic Speech Recognition (ASR) model for transcribing audio.
    """

    @abstractmethod
    async def _transcribe(self, audio: Union[str, io.BytesIO], language: Optional[str] = None) -> str:
        pass

    async def transcribe(self, audio: Union[str, io.BytesIO], language: Optional[str] = None) -> str:
        Tracing.tracers().start_span("asr", inputs={"language": language})
        transcription = await self._transcribe(audio, language)
        Tracing.tracers().end_span(outputs={"transcription": transcription})
        return transcription



from typing import Any, Dict, List, Union


TYPE_MAP = {
    int: 'integer',
    float: 'number',
    bool: 'boolean',
    None: None
}


def is_schema(object_: Any) -> bool:
    return 'type' in object_


def handle_schema(json: Dict[str, Any], schema: Dict[str, Any]) -> None:
    schema.update(json)
    if schema['type'] == 'object':
        schema.pop('properties', None)
        parse(json.get('properties', {}), schema)
    if schema['type'] == 'array':
        schema.pop('items', None)
        schema['items'] = {}
        parse(json.get('items', {}), schema['items'])


def handle_array(arr: List[Any], schema: Dict[str, Any]) -> None:
    schema['type'] = 'array'
    props = schema['items'] = {}
    parse(arr[0], props)


def handle_object(json: Dict[str, Any], schema: Dict[str, Any]) -> None:
    if is_schema(json):
        handle_schema(json, schema)
        return
    schema['type'] = 'object'
    schema['required'] = list(json.keys())
    schema['properties'] = {k: {} for k in json.keys()}
    for key, item in json.items():
        parse(item, schema['properties'][key])


def parse(json: Union[Dict[str, Any], List[Any]], schema: Dict[str, Any]) -> None:
    if isinstance(json, list):
        handle_array(json, schema)
    elif isinstance(json, dict):
        handle_object(json, schema)
    elif isinstance(json, str):
        schema['type'] = 'string'
        schema['description'] = json
    else:
        schema['type'] = TYPE_MAP.get(type(json), 'string')
        schema['example'] = json


def expand_simplified_json_schema(data: Dict[str, Any]) -> Dict[str, Any]:
    json_schema = {}
    parse(data, json_schema)
    return json_schema




from abc import ABC, abstractmethod

from pydantic import BaseModel

from bpm_ai_core.tracing.tracing import Tracing


class QAResult(BaseModel):
    answer: str
    score: float
    start_index: int
    end_index: int


class ExtractiveQA(ABC):
    """
    Extractive Question Answering Model
    """

    @abstractmethod
    def answer_with_metadata(
            self,
            context: str,
            question: str
    ) -> QAResult:
        pass

    def answer(
            self,
            context: str,
            question: str,
            confidence_threshold: float | None = 0.1
    ) -> str:
        Tracing.tracers().start_span("extractive_qa", inputs={
            "context": context,
            "question": question,
            "confidence_threshold": confidence_threshold
        })
        result = self.answer_with_metadata(
            context=context,
            question=question
        )
        Tracing.tracers().end_span(outputs={"result": result.model_dump()})
        # Only return the answer if the score is above the threshold (if given)
        return result.answer \
            if not confidence_threshold or result.score > confidence_threshold \
            else None



import base64
import os
from io import BytesIO

import requests
from PIL import Image

from bpm_ai_core.util.file import is_supported_file

supported_img_extensions = [
    'bmp', 'dib',
    'gif',
    'icns', 'ico',
    'jfif', 'jpe', 'jpeg', 'jpg',
    'j2c', 'j2k', 'jp2', 'jpc', 'jpf', 'jpx',
    'apng', 'png',
    'pbm', 'pgm', 'pnm', 'ppm',
    'tif', 'tiff',
    'webp',
    'emf', 'wmf'
]


def is_supported_img_file(url_or_path: str) -> bool:
    return is_supported_file(url_or_path, supported_extensions=supported_img_extensions)


def load_image(path: str) -> Image:
    """
    Load an image from a local path or a web URL into a Pillow Image object.

    Parameters:
    - path (str): A file system path or a URL of an image.

    Returns:
    - Image: A Pillow Image object.
    """
    if path.startswith('http://') or path.startswith('https://'):
        # Handle web URL
        response = requests.get(path)
        image = Image.open(BytesIO(response.content))
    elif os.path.isfile(path):
        # Handle local file path
        image = Image.open(path)
    else:
        raise ValueError("The path provided is neither a valid URL nor a file path.")

    return image


def base64_encode_image(image: Image):
    """
    Get a base64 encoded string from a Pillow Image object.

    Parameters:
    - image (Image): A Pillow Image object.

    Returns:
    - str: Base64 encoded string of the image.
    """
    buffered = BytesIO()
    image.save(buffered, format=image.format or "JPEG")  # Assuming JPEG if format is not provided
    img_str = base64.b64encode(buffered.getvalue())
    return img_str.decode('utf-8')



import logging

from bpm_ai_core.extractive_qa.question_answering import ExtractiveQA, QAResult

try:
    from transformers import pipeline, AutoTokenizer
    has_transformers = True
except ImportError:
    has_transformers = False

logger = logging.getLogger(__name__)


class TransformersExtractiveQA(ExtractiveQA):
    """
    Local extractive question answering model based on Huggingface transformers library.

    To use, you should have the ``transformers`` python package installed.
    """

    def __init__(self, model: str = "deepset/deberta-v3-large-squad2"):
        if not has_transformers:
            raise ImportError('transformers is not installed')
        self.model = model

    def answer_with_metadata(
            self,
            context: str,
            question: str
    ) -> QAResult:
        qa_model = pipeline("question-answering", model=self.model)

        tokenizer = AutoTokenizer.from_pretrained(self.model)
        tokens = tokenizer.encode(context + question)
        logger.debug(f"Input tokens: {len(tokens)}")

        prediction = qa_model(
            question=question,
            context=context
        )
        logger.debug(f"prediction: {prediction}")

        return QAResult(
            answer=prediction['answer'],
            score=prediction['score'],
            start_index=prediction['start'],
            end_index=prediction['end'],
        )






from abc import ABC, abstractmethod
from typing import Tuple

from pydantic import BaseModel

from bpm_ai_core.tracing.tracing import Tracing


class ClassificationResult(BaseModel):
    max_label: str
    max_score: float
    labels_scores: list[Tuple[str, float]]


class ZeroShotClassifier(ABC):
    """
    Zero Shot Classification Model
    """

    @abstractmethod
    def classify_with_metadata(
            self,
            text: str,
            classes: list[str],
            hypothesis_template: str | None = None
    ) -> ClassificationResult:
        pass

    def classify(
            self,
            text: str,
            classes: list[str],
            confidence_threshold: float | None = None,
            hypothesis_template: str | None = None
    ) -> str:
        Tracing.tracers().start_span("classification", inputs={
            "text": text,
            "classes": classes,
            "confidence_threshold": confidence_threshold,
            "hypothesis_template": hypothesis_template
        })
        result = self.classify_with_metadata(
            text=text,
            classes=classes,
            hypothesis_template=hypothesis_template
        )
        Tracing.tracers().end_span(outputs={"result": result.model_dump()})
        # Only return the label if the score is above the threshold (if given)
        return result.max_label \
            if not confidence_threshold or result.max_score > confidence_threshold \
            else None



from abc import ABC, abstractmethod

from bpm_ai_core.tracing.tracing import Tracing


class NMTModel(ABC):
    """
    Neural Machine Translation Model
    """

    @abstractmethod
    def _translate(self, text: str | list[str], target_language: str) -> str | list[str]:
        pass

    def translate(self, text: str | list[str], target_language: str) -> str | list[str]:
        Tracing.tracers().start_span("nmt", inputs={
            "text": text,
            "target_language": target_language
        })
        translation = self._translate(text, target_language)
        Tracing.tracers().end_span(outputs={"translation": translation})
        return translation



import logging
from typing import Tuple

from bpm_ai_core.pos.pos_tagger import POSTagger

try:
    import spacy
    has_spacy = True
except ImportError:
    has_spacy = False

logger = logging.getLogger(__name__)


def _get_pipeline_for_language(language: str):
    match language:
        case "de":
            return "de_core_news_md"
        case "da":
            return "da_core_news_md"
        case "nl":
            return "nl_core_news_md"
        case "fi":
            return "fi_core_news_md"
        case "fr":
            return "fr_core_news_md"
        case "it":
            return "it_core_news_md"
        case "nn":
            return "nb_core_news_md"
        case "pt":
            return "pt_core_news_md"
        case "es":
            return "es_core_news_md"
        case "sv":
            return "sv_core_news_md"
        case "pl":
            return "pl_core_news_md"
        case "uk":
            return "uk_core_news_md"
        case _:
            return "en_core_web_md"


class SpacyPOSTagger(POSTagger):
    """
    Local POS Tagger based on spaCy library.

    To use, you should have the ``spacy`` python package installed.
    """

    def __init__(self, language: str = "en"):
        if not has_spacy:
            raise ImportError('spacy is not installed')
        pipeline = _get_pipeline_for_language(language)
        for i in range(3):
            try:
                self.nlp = spacy.load(pipeline, disable=["lemmatizer"])
            except OSError:
                from spacy.cli import download
                download(pipeline)
                continue

    def tag(self, text: str) -> list[Tuple[str, str]]:
        doc = self.nlp(text)
        return [(token.text_with_ws, token.pos_) for token in doc]






import io
import os

import requests

from bpm_ai_core.util.file import is_supported_file

supported_audio_extensions = [
   "flac", "mp3", "mp4", "mpeg", "mpga", "m4a", "ogg", "wav2", "webm"
]


def is_supported_audio_file(url_or_path: str) -> bool:
    return is_supported_file(url_or_path, supported_extensions=supported_audio_extensions)


def load_audio(path: str) -> io.BytesIO:
    """
    Load an audio file from a local path or a web URL into a BytesIO object.

    Parameters:
    - path (str): A file system path or a URL of an audio file.

    Returns:
    - BytesIO: A BytesIO object containing the audio data.
    """
    if path.startswith('http://') or path.startswith('https://'):
        # Handle web URL
        response = requests.get(path)
        audio = io.BytesIO(response.content)
    elif os.path.isfile(path):
        # Handle local file path
        with open(path, 'rb') as f:
            audio = io.BytesIO(f.read())
            audio.name = f.name
    else:
        raise ValueError("The path provided is neither a valid URL nor a file path.")

    return audio



import asyncio
import inspect
import json
from typing import Optional, Literal, Any, Union, List

from PIL import Image
from pydantic import BaseModel, Field, ConfigDict

from bpm_ai_core.llm.common.tool import Tool
from bpm_ai_core.tracing.tracing import Tracing


class ChatMessage(BaseModel):
    content: Optional[Union[str, dict, List[Union[str, Image.Image]]]] = None
    """
    The contents of the message. 
    Either a string for normal completions, 
    or a list of strings and images for multimodal completions, 
    or a dict for prediction with output schema.
    """

    role: Literal["system", "user", "assistant", "function"]
    """
    The role of the messages author.
    One of `system`, `user`, `assistant`, or `function`.
    """

    name: Optional[str] = None
    """
    The name of the author of this message.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)


class SingleToolCallMessage(BaseModel):

    id: str

    type: str = Field("function")

    name: str
    """
    The name of the tool that was used.
    """

    payload: Any

    tool: Optional[Tool] = None

    def payload_dict(self) -> dict:
        if isinstance(self.payload, dict):
            return self.payload
        elif isinstance(self.payload, str):
            try:
                return json.loads(self.payload)
            except ValueError as e:
                raise Exception(f"Payload could not be converted to a dict: {e}")
        else:
            raise Exception("Payload has unexpected type.")

    def invoke(self):
        return self.tool.callable(**self.payload_dict())

    async def ainvoke(self) -> Any:
        _callable = self.tool.callable
        inputs = self.payload_dict()
        Tracing.tracers().start_tool_trace(self.tool, inputs)
        if inspect.iscoroutinefunction(_callable):
            result = await _callable(**inputs)
        else:
            result = _callable(**inputs)
        Tracing.tracers().end_tool_trace(result)
        return result


class ToolCallsMessage(ChatMessage):
    role: str = Field("assistant")

    tool_calls: List[SingleToolCallMessage]

    def invoke_all(self) -> List[Any]:
        return [t.invoke() for t in self.tool_calls]

    async def ainvoke_all(self) -> List[Any]:
        return [await t.ainvoke() for t in self.tool_calls]

    async def ainvoke_all_parallel(self) -> List[Any]:
        return await asyncio.gather(*[t.ainvoke() for t in self.tool_calls])


class ToolResultMessage(ChatMessage):

    role: str = Field("tool")

    id: str
    """
    The id of the tool.
    """




import os
from typing import List
from urllib.parse import urlparse


def is_supported_file(url_or_path: str, supported_extensions: List[str]) -> bool:
    url_or_path = url_or_path.strip()
    # Normalize the extensions to lowercase
    supported_extensions = [ext.lower() for ext in supported_extensions]

    # Extract the path from URL if it's a URL
    parsed_url = urlparse(url_or_path)
    path = parsed_url.path if parsed_url.scheme else url_or_path

    # Remove trailing slash if present
    if path.endswith('/'):
        path = path[:-1]

    # Extract the file extension
    _, file_extension = os.path.splitext(path)
    file_extension = file_extension.lower().lstrip('.')

    # Check if the file extension is in the list of supported extensions
    return file_extension in supported_extensions


import json
import logging
import math
import os
import queue
import re
import time
from typing import List, Union, Dict, FrozenSet, Iterable

try:
    import nltk
    import numpy as np
    import torch
    import torch.multiprocessing as mp
    import tqdm
    has_easynmt = True
except ImportError:
    has_easynmt = False

from bpm_ai_core.translation.easy_nmt.opus_mt import OpusMT
from bpm_ai_core.translation.easy_nmt.util import http_get, fullname
from bpm_ai_core.translation.nmt import NMTModel
from bpm_ai_core.util.language import indentify_language

logger = logging.getLogger(__name__)

DOWNLOAD_URL = 'http://easynmt.net/models/v2'


class EasyNMT(NMTModel):
    """
    See https://github.com/UKPLab/EasyNMT.
    Copied and modified to remove dependency to fasttext.
    """

    def __init__(
            self,
            model_name: str = "opus-mt",
            cache_folder: str = None,
            translator=None,
            load_translator: bool = True,
            device=None,
            max_length: int = None,
            **kwargs
    ):
        """
        Easy-to-use, state-of-the-art machine translation
        :param model_name:  Model name (see Readme for available models)
        :param cache_folder: Which folder should be used for caching models. Can also be set via the EASYNMT_CACHE env. variable
        :param translator: Translator object. Set to None, to automatically load the model via the model name.
        :param load_translator: If set to false, it will only load the config but not the translation engine
        :param device: CPU / GPU device for PyTorch
        :param max_length: Max number of token per sentence for translation. Longer text will be truncated
        :param kwargs: Further optional parameters for the different models
        """
        if not has_easynmt:
            raise ImportError('easynmt dependencies is not installed')

        self._model_name = model_name
        self._fasttext_lang_id = None
        self._lang_pairs = frozenset()

        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = device

        self.config = None

        if cache_folder is None:
            if 'EASYNMT_CACHE' in os.environ:
                cache_folder = os.environ['EASYNMT_CACHE']
            else:
                cache_folder = os.path.join(torch.hub._get_torch_home(), 'easynmt_v2')
        self._cache_folder = cache_folder

        if translator is not None:
            self.translator = translator
        else:
            if os.path.exists(model_name) and os.path.isdir(model_name):
                model_path = model_name
            else:
                model_name = model_name.lower()
                model_path = os.path.join(cache_folder, model_name)

                if not os.path.exists(model_path) or not os.listdir(model_path):
                    logger.info("Downloading EasyNMT model {} and saving it at {}".format(model_name, model_path))

                    model_path_tmp = model_path.rstrip("/").rstrip("\\") + "_part"
                    os.makedirs(model_path_tmp, exist_ok=True)

                    # Download easynmt.json
                    config_url = DOWNLOAD_URL + "/{}/easynmt.json".format(model_name)
                    config_path = os.path.join(model_path_tmp, 'easynmt.json')
                    http_get(config_url, config_path)

                    with open(config_path) as fIn:
                        downloaded_config = json.load(fIn)

                    if 'files' in downloaded_config:
                        for filename, url in downloaded_config['files'].items():
                            logger.info("Download {} from {}".format(filename, url))
                            http_get(url, os.path.join(model_path_tmp, filename))

                    ##Rename tmp path
                    try:
                        os.rename(model_path_tmp, model_path)
                    except Exception:
                        pass

            with open(os.path.join(model_path, 'easynmt.json')) as fIn:
                self.config = json.load(fIn)

            if 'lang_pairs' in self.config:
                self._lang_pairs = frozenset(self.config['lang_pairs'])

            if load_translator:
                self.translator = OpusMT(easynmt_path=model_path, **self.config['model_args'])
                self.translator.max_length = max_length

    def _translate(self, text: str | list[str], target_language: str) -> str | list[str]:
        if isinstance(text, str):
            return self.do_translate(text, target_language, indentify_language(text))
        else:
            return [self.do_translate(t, target_language, indentify_language(t)) for t in text]

    def do_translate(
            self,
            documents: Union[str, List[str]],
            target_lang: str,
            source_lang: str,
            show_progress_bar: bool = False,
            beam_size: int = 5,
            batch_size: int = 16,
            perform_sentence_splitting: bool = True,
            paragraph_split: str = "\n",
            sentence_splitter=None,
            **kwargs
    ):
        """
        This method translates the given set of documents
        :param documents: If documents is a string, returns the translated document as string. If documents is a list of strings, translates all documents and returns a list.
        :param target_lang: Target language for the translation
        :param source_lang: Source language for all documents.
        :param show_progress_bar: If true, plot a progress bar on the progress for the translation
        :param beam_size: Size for beam search
        :param batch_size: Number of sentences to translate at the same time
        :param perform_sentence_splitting: Longer documents are broken down sentences, which are translated individually
        :param paragraph_split: Split symbol for paragraphs. No sentences can go across the paragraph_split symbol.
        :param sentence_splitter: Method used to split sentences. If None, uses the default self.sentence_splitting method
        :param kwargs: Optional arguments for the translator model
        :return: Returns a string or a list of string with the translated documents
        """

        # Method_args will store all passed arguments to method
        method_args = locals()
        del method_args['self']
        del method_args['kwargs']
        method_args.update(kwargs)

        if source_lang == target_lang:
            return documents

        is_single_doc = False
        if isinstance(documents, str):
            documents = [documents]
            is_single_doc = True

        if perform_sentence_splitting:
            if sentence_splitter is None:
                sentence_splitter = self.sentence_splitting

            # Split document into sentences
            start_time = time.time()
            splitted_sentences = []
            sent2doc = []
            for doc in documents:
                paragraphs = doc.split(paragraph_split) if paragraph_split is not None else [doc]
                for para in paragraphs:
                    for sent in sentence_splitter(para.strip(), source_lang):
                        sent = sent.strip()
                        if len(sent) > 0:
                            splitted_sentences.append(sent)
                sent2doc.append(len(splitted_sentences))
            # logger.info("Sentence splitting done after: {:.2f} sec".format(time.time() - start_time))
            # logger.info("Translate {} sentences".format(len(splitted_sentences)))

            translated_sentences = self.translate_sentences(splitted_sentences, target_lang=target_lang,
                                                            source_lang=source_lang,
                                                            show_progress_bar=show_progress_bar, beam_size=beam_size,
                                                            batch_size=batch_size, **kwargs)

            # Merge sentences back to documents
            start_time = time.time()
            translated_docs = []
            for doc_idx in range(len(documents)):
                start_idx = sent2doc[doc_idx - 1] if doc_idx > 0 else 0
                end_idx = sent2doc[doc_idx]
                translated_docs.append(
                    self._reconstruct_document(documents[doc_idx], splitted_sentences[start_idx:end_idx],
                                               translated_sentences[start_idx:end_idx]))

            # logger.info("Document reconstruction done after: {:.2f} sec".format(time.time() - start_time))
        else:
            translated_docs = self.translate_sentences(documents, target_lang=target_lang, source_lang=source_lang,
                                                       show_progress_bar=show_progress_bar, beam_size=beam_size,
                                                       batch_size=batch_size, **kwargs)

        if is_single_doc:
            translated_docs = translated_docs[0]

        return translated_docs

    @staticmethod
    def _reconstruct_document(doc, org_sent, translated_sent):
        """
        This method reconstructs the translated document and
        keeps white space in the beginning / at the end of sentences.
        """
        sent_idx = 0
        char_idx = 0
        translated_doc = ""
        while char_idx < len(doc):
            if sent_idx < len(org_sent) and doc[char_idx] == org_sent[sent_idx][0]:
                translated_doc += translated_sent[sent_idx]
                char_idx += len(org_sent[sent_idx])
                sent_idx += 1
            else:
                translated_doc += doc[char_idx]
                char_idx += 1
        return translated_doc

    def translate_sentences(
            self,
            sentences: Union[str, List[str]],
            target_lang: str,
            source_lang: str,
            show_progress_bar: bool = False,
            beam_size: int = 5,
            batch_size: int = 32,
            **kwargs
    ):
        """
        This method translates individual sentences.

        :param sentences: A single sentence or a list of sentences to be translated
        :param source_lang: Source language for all sentences.
        :param target_lang: Target language for the translation
        :param show_progress_bar: Show a progress bar
        :param beam_size: Size for beam search
        :param batch_size: Mini batch size
        :return: List of translated sentences
        """

        if source_lang == target_lang:
            return sentences

        is_single_sentence = False
        if isinstance(sentences, str):
            sentences = [sentences]
            is_single_sentence = True

        output = []

        # Sort by length to speed up processing
        length_sorted_idx = np.argsort([-len(sen) for sen in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]

        iterator = range(0, len(sentences_sorted), batch_size)
        if show_progress_bar:
            scale = min(batch_size, len(sentences))
            iterator = tqdm.tqdm(iterator, total=len(sentences) / scale, unit_scale=scale, smoothing=0)

        for start_idx in iterator:
            output.extend(self.translator.translate_sentences(sentences_sorted[start_idx:start_idx + batch_size],
                                                              source_lang=source_lang, target_lang=target_lang,
                                                              beam_size=beam_size, device=self.device, **kwargs))

        # Restore original sorting of sentences
        output = [output[idx] for idx in np.argsort(length_sorted_idx)]

        if is_single_sentence:
            output = output[0]

        return output

    def start_multi_process_pool(self, target_devices: List[str] = None):
        """
        Starts multi process to process the encoding with several, independent processes.
        This method is recommended if you want to encode on multiple GPUs. It is advised
        to start only one process per GPU. This method works together with encode_multi_process
        :param target_devices: PyTorch target devices, e.g. cuda:0, cuda:1... If None, all available CUDA devices will be used
        :return: Returns a dict with the target processes, an input queue and and output queue.
        """
        if target_devices is None:
            if torch.cuda.is_available():
                target_devices = ['cuda:{}'.format(i) for i in range(torch.cuda.device_count())]
            else:
                logger.info("CUDA is not available. Start 4 CPU worker")
                target_devices = ['cpu'] * 4

        logger.info("Start multi-process pool on devices: {}".format(', '.join(map(str, target_devices))))

        ctx = mp.get_context('spawn')
        input_queue = ctx.Queue()
        output_queue = ctx.Queue()
        processes = []

        for cuda_id in target_devices:
            p = ctx.Process(target=EasyNMT._encode_multi_process_worker,
                            args=(cuda_id, self, input_queue, output_queue), daemon=True)
            p.start()
            processes.append(p)

        return {'input': input_queue, 'output': output_queue, 'processes': processes}

    def translate_multi_process(
            self,
            pool: Dict[str, object],
            documents: List[str],
            show_progress_bar: bool = True,
            chunk_size: int = None, **kwargs
    ) -> List[str]:
        """
        This method allows to run encode() on multiple GPUs. The sentences are chunked into smaller packages
        and sent to individual processes, which encode these on the different GPUs. This method is only suitable
        for encoding large sets of sentences
        :param pool: A pool of workers started with SentenceTransformer.start_multi_process_pool
        :param chunk_size: Sentences are chunked and sent to the individual processes. If none, it determine a sensible size.
        """

        if chunk_size is None:
            chunk_size = min(math.ceil(len(documents) / len(pool["processes"]) / 10), 1000)

        logger.info("Chunk data into packages of size {}".format(chunk_size))

        input_queue = pool['input']
        last_chunk_id = 0

        for start_idx in range(0, len(documents), chunk_size):
            input_queue.put([last_chunk_id, documents[start_idx:start_idx + chunk_size], kwargs])
            last_chunk_id += 1

        output_queue = pool['output']
        results_list = sorted([output_queue.get() for _ in
                               tqdm.tqdm(range(last_chunk_id), total=last_chunk_id, unit_scale=chunk_size, smoothing=0,
                                         disable=not show_progress_bar)], key=lambda chunk: chunk[0])
        translated = []
        for chunk in results_list:
            translated.extend(chunk[1])
        return translated

    def translate_stream(self, stream: Iterable[str], show_progress_bar: bool = True, chunk_size: int = 128,
                         **kwargs) -> List[str]:
        batch = []
        for doc in tqdm.tqdm(stream, smoothing=0.0, disable=not show_progress_bar):
            batch.append(doc)

            if len(batch) >= chunk_size:
                translated = self.do_translate(batch, show_progress_bar=False, **kwargs)
                for trans_doc in translated:
                    yield trans_doc
                batch = []

        if len(batch) > 0:
            translated = self.do_translate(batch, show_progress_bar=False, **kwargs)
            for trans_doc in translated:
                yield trans_doc

    @staticmethod
    def stop_multi_process_pool(pool):
        """
        Stops all processes started with start_multi_process_pool
        """
        for p in pool['processes']:
            p.terminate()

        for p in pool['processes']:
            p.join()
            p.close()

        pool['input'].close()
        pool['output'].close()

    @staticmethod
    def _encode_multi_process_worker(target_device: str, model, input_queue, results_queue):
        """
        Internal working process to encode sentences in multi-process setup
        """
        model.device = target_device
        while True:
            try:
                id, documents, kwargs = input_queue.get()
                translated = model.do_translate(documents, **kwargs)
                results_queue.put([id, translated])
            except queue.Empty:
                break

    def sentence_splitting(self, text: str, lang: str = None):
        if lang in ['ar', 'jp', 'ko', 'zh']:
            sentences = list(re.findall(u'[^!?。\.]+[!?。\.]*', text, flags=re.U))
        else:
            try:
                nltk.data.find('tokenizers/punkt')
            except LookupError:
                nltk.download('punkt')

            sentences = nltk.sent_tokenize(text)

        return sentences

    @property
    def lang_pairs(self) -> FrozenSet[str]:
        """
        Returns all allowed languages directions for the loaded model
        """
        return self._lang_pairs

    def get_languages(self, source_lang: str = None, target_lang: str = None) -> List[str]:
        """
        Returns all available languages supported by the model
        :param source_lang:  If not None, then returns all languages to which we can translate for the given source_lang
        :param target_lang:  If not None, then returns all languages from which we can translate for the given target_lang
        :return: Sorted list with the determined languages
        """

        langs = set()
        for lang_pair in self.lang_pairs:
            source, target = lang_pair.split("-")

            if source_lang is None and target_lang is None:
                langs.add(source)
                langs.add(target)
            elif target_lang is not None and target == target_lang:
                langs.add(source)
            elif source_lang is not None and source == source_lang:
                langs.add(target)

        return sorted(list(langs))

    def save(self, output_path):
        os.makedirs(output_path, exist_ok=True)

        filepath = os.path.join(output_path, 'easynmt.json')

        config = {
            'model_class': fullname(self.translator),
            'lang_pairs': list(self.lang_pairs),
            'model_args': self.translator.save(output_path)
        }

        with open(filepath, 'w') as fOut:
            json.dump(config, fOut)



import json
from typing import Dict, Any, Optional, List

from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.llm.common.message import ChatMessage, ToolCallsMessage, SingleToolCallMessage
from bpm_ai_core.llm.common.tool import Tool
from bpm_ai_core.util.openai import messages_to_openai_dicts, json_schema_to_openai_function

logger = logging.getLogger(__name__)

try:
    from openai import AsyncOpenAI, APIConnectionError, InternalServerError, RateLimitError, OpenAIError
    from openai.types.chat import ChatCompletionMessage, ChatCompletion
    import httpx

    has_openai = True
    try:
        client = AsyncOpenAI(
            http_client=httpx.AsyncClient(
                limits=httpx.Limits(max_connections=1000, max_keepalive_connections=100)
            ),
            max_retries=0  # we use own retry logic
        )
    except OpenAIError as e:
        logger.error(e)
except ImportError:
    has_openai = False


class ChatOpenAI(LLM):
    """
    `OpenAI` Chat large language models API.

    To use, you should have the ``openai`` python package installed, and the
    environment variable ``OPENAI_API_KEY`` set with your API key.
    """

    def __init__(
        self,
        model: str = "gpt-3.5-turbo-0125",
        temperature: float = 0.0,
        seed: Optional[int] = None,
        max_retries: int = 8
    ):
        if not has_openai:
            raise ImportError('openai is not installed')
        super().__init__(
            model=model,
            temperature=temperature,
            max_retries=max_retries,
            retryable_exceptions=[
                RateLimitError, InternalServerError, APIConnectionError
            ]
        )
        self.seed = seed

    async def _predict(
        self,
        messages: List[ChatMessage],
        output_schema: Optional[Dict[str, Any]] = None,
        tools: Optional[List[Tool]] = None
    ) -> ChatMessage:
        openai_tools = []
        if output_schema:
            tools = [Tool.from_callable(name="store_result", description="Stores your result", args_schema=output_schema)]
        if tools:
            openai_tools = [json_schema_to_openai_function(f.name, f.description, f.args_schema) for f in tools]
        completion = await self._run_completion(messages, openai_tools)

        message = completion.choices[0].message
        if message.tool_calls:
            if output_schema:
                return ChatMessage(role=message.role, content=self._load_tool_call_json(message))
            else:
                return self._openai_tool_calls_to_tool_message(message, tools)
        else:
            return ChatMessage(role=message.role, content=message.content)

    async def _run_completion(self, messages: List[ChatMessage], functions: List[dict]) -> ChatCompletion:
        args = {
            "model": self.model,
            "temperature": self.temperature,
            **({"seed": self.seed} if self.seed else {}),
            "messages": messages_to_openai_dicts(messages),
            **({
                   "tool_choice": {
                       "type": "function",
                       "function": {"name": functions[0]["function"]["name"]}
                   } if (len(functions) == 1) else ("auto" if functions else "none"),
                   "tools": functions
               } if functions else {})
        }
        return await client.chat.completions.create(**args)

    @staticmethod
    def _openai_tool_calls_to_tool_message(message: ChatCompletionMessage, tools: List[Tool]) -> ToolCallsMessage:
        return ToolCallsMessage(
            name=", ".join([t.function.name for t in message.tool_calls]),
            content=message.content,
            tool_calls=[
                SingleToolCallMessage(
                    id=t.id,
                    name=t.function.name,
                    payload=t.function.arguments,
                    tool=next((item for item in tools if item.name == t.function.name), None)
                )
                for t in message.tool_calls
            ]
        )

    @staticmethod
    def _load_tool_call_json(message: ChatCompletionMessage):
        try:
            json_object = json.loads(message.tool_calls[0].function.arguments)
        except ValueError as e:
            json_object = None
        return json_object

    def name(self) -> str:
        return "openai"



import logging
import time
from typing import List

try:
    import torch
    from transformers import MarianMTModel, MarianTokenizer
    has_transformers = True
except ImportError:
    has_transformers = False

logger = logging.getLogger(__name__)


class OpusMT:
    def __init__(self, easynmt_path: str = None, max_loaded_models: int = 10):
        if not has_transformers:
            raise ImportError('transformers is not installed')
        self.models = {}
        self.max_loaded_models = max_loaded_models
        self.max_length = None

    def load_model(self, model_name):
        if model_name in self.models:
            self.models[model_name]['last_loaded'] = time.time()
            return self.models[model_name]['tokenizer'], self.models[model_name]['model']
        else:
            logger.info("Load model: "+model_name)
            tokenizer = MarianTokenizer.from_pretrained(model_name)
            model = MarianMTModel.from_pretrained(model_name)
            model.eval()

            if len(self.models) >= self.max_loaded_models:
                oldest_time = time.time()
                oldest_model = None
                for loaded_model_name in self.models:
                    if self.models[loaded_model_name]['last_loaded'] <= oldest_time:
                        oldest_model = loaded_model_name
                        oldest_time = self.models[loaded_model_name]['last_loaded']
                del self.models[oldest_model]

            self.models[model_name] = {'tokenizer': tokenizer, 'model': model, 'last_loaded': time.time()}
            return tokenizer, model

    def translate_sentences(self, sentences: List[str], source_lang: str, target_lang: str, device: str, beam_size: int = 5, **kwargs):
        model_name = 'Helsinki-NLP/opus-mt-{}-{}'.format(source_lang, target_lang)
        tokenizer, model = self.load_model(model_name)
        model.to(device)

        inputs = tokenizer(sentences, truncation=True, padding=True, max_length=self.max_length, return_tensors="pt")

        for key in inputs:
            inputs[key] = inputs[key].to(device)

        with torch.no_grad():
            translated = model.generate(**inputs, num_beams=beam_size, **kwargs)
            output = [tokenizer.decode(t, skip_special_tokens=True) for t in translated]

        return output

    def save(self, output_path):
        return {"max_loaded_models": self.max_loaded_models}




import logging

from bpm_ai_core.classification.zero_shot_classifier import ZeroShotClassifier, ClassificationResult

try:
    from transformers import pipeline, AutoTokenizer
    has_transformers = True
except ImportError:
    has_transformers = False

logger = logging.getLogger(__name__)

DEFAULT_MODEL_EN = "MoritzLaurer/deberta-v3-large-zeroshot-v1.1-all-33"
DEFAULT_MODEL_MULTI = "MoritzLaurer/mDeBERTa-v3-base-mnli-xnli"


class TransformersClassifier(ZeroShotClassifier):
    """
    Local zero-shot classification model based on Huggingface transformers library.

    To use, you should have the ``transformers`` python package installed.
    """

    def __init__(self, model: str = DEFAULT_MODEL_EN):
        if not has_transformers:
            raise ImportError('transformers is not installed')
        self.model = model

    def classify_with_metadata(
            self,
            text: str,
            classes: list[str],
            hypothesis_template: str | None = None
    ) -> ClassificationResult:
        zeroshot_classifier = pipeline("zero-shot-classification", model=self.model)

        tokenizer = AutoTokenizer.from_pretrained(self.model)
        input_tokens = len(tokenizer.encode(text))
        max_tokens = tokenizer.model_max_length
        logger.debug(f"Input tokens: {input_tokens}")
        if input_tokens > max_tokens:
            logger.warning(
                f"Input tokens exceed max model context size: {input_tokens} > {max_tokens}. Input will be truncated."
            )

        prediction = zeroshot_classifier(
            text,
            classes,
            hypothesis_template=hypothesis_template or "This example is about {}",
            multi_label=False
        )
        # Zip the labels and scores together and find the label with the max score
        labels_scores = list(zip(prediction['labels'], prediction['scores']))
        max_label, max_score = max(labels_scores, key=lambda x: x[1])

        return ClassificationResult(
            max_label=max_label,
            max_score=max_score,
            labels_scores=labels_scores
        )






import os
import requests
import sys

def http_get(url, path):
    """
    Downloads a URL to a given path on disc
    """
    if os.path.dirname(path) != '':
        os.makedirs(os.path.dirname(path), exist_ok=True)

    req = requests.get(url, stream=True)
    if req.status_code != 200:
        print("Exception when trying to download {}. Response {}".format(url, req.status_code), file=sys.stderr)
        req.raise_for_status()
        return

    download_filepath = path+"_part"
    with open(download_filepath, "wb") as file_binary:
        content_length = req.headers.get('Content-Length')
        total = int(content_length) if content_length is not None else None
        import tqdm
        progress = tqdm.tqdm(unit="B", total=total, unit_scale=True)
        for chunk in req.iter_content(chunk_size=1024):
            if chunk: # filter out keep-alive new chunks
                progress.update(len(chunk))
                file_binary.write(chunk)

    os.rename(download_filepath, path)
    progress.close()

def fullname(o):
  """
  Gives a full name (package_name.class_name) for a class / object in Python. Will
  be used to load the correct classes from JSON files
  """

  module = o.__class__.__module__
  if module is None or module == str.__class__.__module__:
    return o.__class__.__name__  # Avoid reporting __builtin__
  else:
    return module + '.' + o.__class__.__name__

def fullname(o):
  """
  Gives a full name (package_name.class_name) for a class / object in Python. Will
  be used to load the correct classes from JSON files
  """

  module = o.__class__.__module__
  if module is None or module == str.__class__.__module__:
    return o.__class__.__name__  # Avoid reporting __builtin__
  else:
    return module + '.' + o.__class__.__name__


from typing import Dict, Any, Optional, List

from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.llm.common.message import ChatMessage
from bpm_ai_core.llm.common.tool import Tool

try:
    from anthropic import Anthropic, HUMAN_PROMPT, AI_PROMPT
    has_anthropic = True
except ImportError:
    has_anthropic = False


class AnthropicChat(LLM):

    def __init__(
        self,
        model: str = "claude-2.1",
        temperature: float = 0.0,
        max_retries: int = 8,
        client_kwargs: Optional[Dict[str, Any]] = None
    ):
        if not has_anthropic:
            raise ImportError('anthropic is not installed')
        super().__init__(
            model=model,
            temperature=temperature,
            max_retries=max_retries
        )
        self.client = Anthropic(
            max_retries=0,  # we use own retry logic
            **(client_kwargs or {})
        )

    def _predict(
        self,
        messages: List[ChatMessage],
        output_schema: Optional[Dict[str, Any]] = None,
        tools: Optional[List[Tool]] = None
    ) -> ChatMessage:

        completion = self._run_completion(messages, [])

        return ChatMessage(
            role="assistant",
            content=completion.completion.strip()
        )

    def _run_completion(self, messages: List[ChatMessage], functions: List[dict]):
        prompt = AnthropicChat.messages_to_claude_str(messages)
        return self.client.completions.create(
            max_tokens_to_sample=4096,
            model=self.model,
            temperature=self.temperature,
            prompt=prompt
        )

    @staticmethod
    def messages_to_claude_str(messages: List[ChatMessage]) -> str:
        return "".join([AnthropicChat.message_to_claude_str(m) for m in messages] + [AI_PROMPT])

    @staticmethod
    def message_to_claude_str(m) -> str:
        if m.role == "assistant":
            header = AI_PROMPT + " "
        elif m.role == "user":
            header = HUMAN_PROMPT + " "
        else:
            header = ""
        return header + m.content

    def name(self) -> str:
        return "anthropic"
